import statsmodels.api as sm
import pandas as pd
import numpy as np
from numpy import nan
from multiprocessing import Pool
from functools import partial
import time

from .pdutils import split

def reg_by(df, yvar, xvars, groupvar, merge=False, cons=True, mp=False):
    """
    Runs a regression of df[yvar] on df[xvars] by values of groupvar. Outputs a dataframe with values of
    groupvar and corresponding coefficients, unless merge=True, then outputs the original dataframe with the
    appropriate coefficients merged in.

    Required inputs:
    groupvar: str or list of strs, column names of columns identifying by groups

    Optional Options:
    cons: True to include a constant, False to not
    mp: False to use single processor, True to use all processors, int to use # processors
    """
    # Convert xvars to list if str is passed
    xvars = _check_inputs_regby(xvars)

    # If there are multiple groupvars, create a key variable containing all groupvars (modifies df inplace)
    groupvar, drop_group = _set_groupvar_and_drop_group(df, groupvar)

    # Select dataframe of only y, x and group vars, drop any missings
    yx_df = df_for_reg(df, yvar, xvars, groupvar)

    # Create a list of right hand side variables. Includes 'const' if cons is True
    rhs = _set_rhs(xvars, cons)

    # Split DataFrame into a list of arrays with each bygroup being one array. Provide an accompanying list of bygroups
    arrs, groups = _get_lists_of_arrays_and_groups(yx_df, yvar, xvars, groupvar)

    # Run regressions by groups, storing results as a list of numpy arrays
    results = _reg_by(arrs, groups, xvars, rhs, cons, mp=mp)

    # Combine list of arrays into df, and apply column labels
    result_df = _result_list_of_arrays_to_df(results, rhs, groupvar)

    if merge:
        result_df = df.merge(result_df, how='left', on=groupvar)
    if drop_group:
        result_df.drop(groupvar, axis=1, inplace=True)

    return result_df.reset_index(drop=True)

def df_for_reg(df, yvar, xvars, groupvar):
    # Select dataframe of only y and x vars
    yx_df = df.loc[:, xvars + [yvar]]
    # Recombine groupvar and drop missing
    yx_df = pd.concat([yx_df, df[groupvar]], axis=1).dropna()

    return yx_df

def _reg_by(arrs, groups, xvars, rhs, cons, mp=False):
    if mp:
        return _reg_by_mp(arrs, groups, xvars, rhs, cons)
    else:
        return _reg_by_sp(arrs, groups, xvars, rhs, cons)

def _reg_by_sp(arrs, groups, xvars, rhs, cons):
    results = []
    for i, arr in enumerate(arrs):
        results.append(_reg(arr, xvars, rhs, cons, groups[i]))
    return results

def _reg_by_mp(arrs, groups, xvars, rhs, cons, mp=True):
    if isinstance(mp, int):
        num_processes = mp
    else:
        num_processes = None

    # result_counter = ResultCounter(len(arrs))
    # _log_with_result_counter = partial(
    #     _log_mp_status,
    #     result_counter=result_counter
    # )
    num_expected_results = len(arrs)
    num_completed = 0

    with Pool(num_processes) as pool:
        expected_results = []
        for i, arr in enumerate(arrs):
            expected_results.append(
                pool.apply_async(
                    _reg,
                    (arr, xvars, rhs, cons, groups[i])
                    # callback=_log_with_result_counter
                )
            )

        while num_completed != num_expected_results:
            time.sleep(1)
            num_completed = sum([1 for result in expected_results if result.ready()])
            print(f'Finished {num_completed} out of {num_expected_results} calculations', end='\r')

        completed_results = [result.get() for result in expected_results]

    return completed_results


def _log_mp_status(result, result_counter):
    result_counter += 1
    print(result_counter, end='\r')

class ResultCounter:

    def __init__(self, num_expected_results):
        self.count = 0
        self.expected_results = num_expected_results

    def __add__(self, other):
        self.count += other

    def __repr__(self):
        return rf'<ResultCounter(count={self.count}, expected={self.expected_results})>'

    def __str__(self):
        return f'Finished {completed_results} out of {num_expected_results} calculations'

def _reg(arr, xvars, rhs, cons, group):
    X = arr[:, 1:].astype(float)

    if cons:
        X = sm.add_constant(X)

    y = arr[:, 0].astype(float)

    if arr.shape[0] > len(xvars) + 1:  # if enough observations, run regression
        model = sm.OLS(y, X)
        result = model.fit()
        this_result = np.append(result.params, group)  # add groupvar
        this_result = this_result[None, :]  # cast 1d array into 2d array
    else:  # not enough obs, return nans
        this_result = np.empty((1, len(rhs) + 1), dtype='O')
        this_result[:] = nan
        this_result[0, len(rhs)] = group

    return this_result


def _check_inputs_regby(xvars):
    if isinstance(xvars, str):
        xvars = [xvars]
    assert isinstance(xvars, list)

    return xvars

def _set_groupvar_and_drop_group(df, groupvar):
    drop_group = False
    if isinstance(groupvar, list):
        df['__key_regby__'] = ''
        for var in groupvar:
            df['__key_regby__'] = df['__key_regby__'] + df[var].astype(str)
        groupvar = '__key_regby__'
        drop_group = True

    return groupvar, drop_group

def _set_rhs(xvars, cons):
    if cons:
        rhs = ['const'] + xvars
    else:
        rhs = xvars

    return rhs

def _result_list_of_arrays_to_df(results, rhs, groupvar):
    result_df = pd.DataFrame(np.concatenate(results, axis=0))
    result_df = result_df.apply(pd.to_numeric, errors='ignore')
    cols = rhs + [groupvar]
    result_df.columns = ['coef_' + col if col not in (groupvar, 'const') else col for col in cols]

    return result_df

def _get_lists_of_arrays_and_groups(df, yvar, xvars, groupvar):
    df.sort_values(groupvar, inplace=True)
    arrs = split(df, [yvar] + xvars, keyvar=groupvar)
    groups = df[groupvar].unique().tolist()
    assert len(arrs) == len(groups)

    return arrs, groups



import os
from functools import partial

from .dropbox import filepath as dropbox_filepath

data_filepath = r'E:\Data'

def data_path(subpath):
    return os.path.join(data_filepath, subpath)

def data_path_func(subpath):
    return path_func(subpath, data_path)

def dropbox_path(subpath):
    return os.path.join(dropbox_filepath, subpath)

def dropbox_path_func(subpath):
    return path_func(subpath, dropbox_path)

def path_func(subpath, path_func):
    new_path = path_func(subpath)
    new_path_func = partial(os.path.join, new_path)
    return new_path_func

#### Project paths

etf_project_path = dropbox_path_func('UF\Andy\ETF Project')
etf_path_func = lambda subpath: path_func(subpath, etf_project_path)



class RegressionSetArgs:
    def __init__(self, df, yvar, xvars_list, fe_list=None, **reg_kwargs):
        self.df = df
        self.yvar = yvar
        self.xvars_list = xvars_list
        self.fe = fe_list
        self._parse_reg_kwargs(reg_kwargs)

    def keys(self):
        return ['df', 'yvar', 'xvars_list', 'fe'] + self.reg_kwarg_names

    def _parse_reg_kwargs(self, reg_kwargs):
        self.reg_kwarg_names = []

        for name, arg in reg_kwargs.items():
            # for each arg in reg_kwargs, add as class attribute
            setattr(self, name, arg)
            self.reg_kwarg_names.append(name) #also add to list of names for keys()

    def __getitem__(self, item):
        return getattr(self, item)


from dero.ext_pandas import _to_list_if_str
from dero.ext_pandas.filldata import add_missing_group_rows, drop_missing_group_rows


def create_lagged_variables(df, lag_cols, id_col='TICKER', date_col='Date', num_lags=1):
    """
    Note: partially inplace
    """
    df.sort_values([id_col, date_col], inplace=True)
    df = add_missing_group_rows(df, [id_col, date_col])

    for col in lag_cols:
        _create_lagged_variable(df, col, id_col=id_col, num_lags=num_lags)

    df = drop_missing_group_rows(df, [id_col, date_col])

    return df


def _create_lagged_variable(df, col, id_col='TICKER', num_lags=1):
    """
    Note: inplace
    """
    new_name = varname_to_lagged_varname(col, num_lags=num_lags)
    df[new_name] = df.groupby(id_col)[col].shift(num_lags)


def varname_to_lagged_varname(varname, num_lags=1):
    return varname + f'_{{t - {num_lags}}}'


def _convert_variable_names(yvar, xvars, lag_cols, num_lags=1):
    if yvar in lag_cols:
        yvar = varname_to_lagged_varname(yvar, num_lags=num_lags)

    out_xvars = []
    for xvar in xvars:
        if xvar in lag_cols:
            out_xvars.append(varname_to_lagged_varname(xvar, num_lags=num_lags))
        else:
            out_xvars.append(xvar)

    return yvar, out_xvars


def _convert_interaction_tuples(interaction_tuples, lag_cols, num_lags=1):
    out_tuples = []
    for tup in interaction_tuples:
        out_tuples.append(
            tuple([
                varname_to_lagged_varname(var, num_lags=num_lags)
                if (var in lag_cols) or (var + ' Change' in lag_cols)
                else var
                for var in tup
            ])
        )
    return out_tuples


def _set_lag_variables(lag_variables, yvar, xvars):
    # Already passing a collection of columns, return
    if isinstance(lag_variables, (list, tuple)):
        return lag_variables

    assert isinstance(lag_variables, str)

    #Single str can either be a single column, 'all', or 'xvars'
    if lag_variables == 'xvars':
        lag_variables = xvars.copy()
    elif lag_variables == 'all':
        lag_variables = [yvar] + xvars
    else: #single column passed
        return _to_list_if_str(lag_variables)

    return lag_variables


def _is_special_lag_keyword(lag_variables):
    if isinstance(lag_variables, (list, tuple)):
        return False #list of columns

    special_keywords = ('xvars', 'all')
    return lag_variables in special_keywords


from dero.latex.models.mixins import StringAdditionMixin
from dero.latex.texgen import _include_graphics_str


class Graphic(StringAdditionMixin):

    def __init__(self, filepath, width=r'\linewidth'):
        self.filepath = filepath
        self.width = width

    def __repr__(self):
        return f'<Graphic({self.filepath}, width={self.width})>'

    def __str__(self):
        return _include_graphics_str(self.filepath, self.width)





import pandas as pd

class SimplifiedRegressionResult:
    def __init__(self, params, pvalues, nobs, rsquared_adj, tvalues, bse, conf_int, exog_names, endog_names):
        self.params = params
        self.pvalues = pvalues
        self.nobs = nobs
        self.rsquared_adj = rsquared_adj
        self.tvalues = tvalues
        self.bse = bse
        self.conf_int = conf_int
        self.model = SimplifiedModel(exog_names, endog_names)

    @classmethod
    def from_statsmodels_result(cls, result):

        # Get direct attributes
        pull_attrs = ['params', 'pvalues', 'tvalues', 'nobs', 'rsquared_adj', 'bse', 'conf_int']
        result_dict = _extract_attrs_into_dict(result, pull_attrs)

        # Get attributes of model
        model_attrs = ['exog_names', 'endog_names']
        result_dict.update(_extract_attrs_into_dict(result.model, model_attrs))

        return cls(**result_dict)

class SimplifiedModel:

    def __init__(self, exog_names, endog_names):
        self.exog_names = exog_names
        self.endog_names = endog_names

def _extract_attrs_into_dict(obj, attrs):
    result_dict = {}

    # Get direct attributes
    for attr in attrs:
        value = getattr(obj, attr)
        if isinstance(value, (pd.Series, pd.DataFrame, list, dict)):
            value = value.copy()
        result_dict[attr] = value

    return result_dict


def remove_lag_names_from_reg_results(reg_list, lags=(1,)):
    """
    Note: partially inplace
    """
    out_reg_list = []
    for ambiguous_result in reg_list:

        # Determine type of result
        if isinstance(ambiguous_result, tuple):  # Tuple of result, fe_dict
            result = ambiguous_result[0]
        else:  # just a single result
            result = ambiguous_result

        # Actually replace names
        result = _remove_lag_name_from_reg_result(result, lags=lags)

        # Add to output, depending on type of result
        if isinstance(ambiguous_result, tuple):  # Tuple of result, fe_dict
            out_reg_list.append((result, ambiguous_result[1]))
        else:  # just a single result
            out_reg_list.append(result)

    return out_reg_list


def _remove_lag_name_from_reg_result(result, lags=(1,)):
    """
    Note: partially inplace
    """
    result = SimplifiedRegressionResult.from_statsmodels_result(result)

    # Modify base properties inplace
    [_remove_lag_names_from_series_index(getattr(result, item), lags=lags) for item in ('params', 'pvalues', 'tvalues', 'bse')]

    # Modify model properties and reassign (functions not inplace)
    result.model.endog_names = _remove_lag_names_from_varname(result.model.endog_names, lags=lags)
    result.model.exog_names = _remove_lag_names_from_list(result.model.exog_names, lags=lags)

    return result


def _remove_lag_names_from_series_index(series, lags=(1,)):
    """
    Note: inplace
    """
    [_remove_one_lag_names_from_series_index(series, num_lags=num_lags) for num_lags in lags]


def _remove_one_lag_names_from_series_index(series, num_lags=1):
    """
    Note: inplace
    """
    rename_dict = {col: lag_varname_to_varname(col, num_lags=num_lags) for col in series.index}
    series.index = series.index.to_series().replace(rename_dict)

def _remove_lag_names_from_list(list_, lags=(1,)):
    for lag in lags:
        list_ = _remove_one_lag_names_from_list(list_, lag)
    return list_

def _remove_one_lag_names_from_list(list_, num_lags=1):
    return [lag_varname_to_varname(item, num_lags=num_lags) for item in list_]

def _remove_lag_names_from_varname(varname, lags=(1,)):
    for lag in lags:
        varname = lag_varname_to_varname(varname, lag)
    return varname

def lag_varname_to_varname(varname, num_lags=1):
    return varname.replace(rf'_{{t - {num_lags}}}', '')



import os
import posixpath
import shutil

from dero.latex.figure.models import Subfigure, Figure
from dero.latex.models import Document, Package
from dero.latex.tools import date_time_move_latex

def filepaths_to_pdf_figure_and_move(filepaths,  outfolder, outname='figure', as_document=True,
                                     subfigure_kwargs={}, figure_kwargs={}, document_kwargs={}):

    basenames = [_latex_valid_basename(filepath) for filepath in filepaths]
    sources_paths = [posixpath.join('Sources', basename) for basename in basenames]

    document = _filepaths_to_document(
        sources_paths,
        subfigure_kwargs=subfigure_kwargs,
        figure_kwargs=figure_kwargs,
        document_kwargs=document_kwargs
    )

    _document_to_pdf_and_move(
        document,
        outfolder,
        image_paths=filepaths,
        outname=outname,
        as_document=as_document
    )

    return document

def _document_to_pdf_and_move(document, outfolder, image_paths=None, outname='figure', as_document=True):

    # We will change paths, so save original to switch back to
    orig_path = os.getcwd()

    os.chdir(outfolder)

    # Create tex file
    outname_tex = outname + '.tex'
    with open(outname_tex, 'w') as f:
        f.write(str(document))

    if image_paths:
        # Copy first time for creation of pdf
        sources_tempfolder = os.path.join(outfolder, 'Sources')
        if not os.path.exists(sources_tempfolder):
            os.makedirs(sources_tempfolder)
        [_copy_if_needed(filepath, os.path.join(sources_tempfolder, _latex_valid_basename(filepath)))
         for filepath in image_paths]

    if as_document:
        os.system('pdflatex ' + '"' + outname_tex + '"') #create PDF
    new_outfolder = date_time_move_latex(outname, outfolder, folder_name='Figures') #move table into appropriate date/number folder
    sources_outfolder = os.path.join(new_outfolder, 'Sources')

    if image_paths and new_outfolder:
        # Copy second time to move pictures along with pdf
        _move_if_needed(sources_tempfolder, sources_outfolder)

    os.chdir(orig_path)

def _filepaths_to_document(filepaths, subfigure_kwargs={}, figure_kwargs={}, document_kwargs={}):
    subfigures = [Subfigure(fp, **subfigure_kwargs) for fp in filepaths]
    figure = Figure(subfigures, **figure_kwargs)

    simple_package_strs = [
        'caption',
        'subcaption',
        'graphicx',
        'pdflscape'
    ]
    simple_packages = [Package(str_) for str_ in simple_package_strs]

    packages = simple_packages + [
        Package('geometry', modifier_str='margin=0.1in')
    ]

    document = Document(
        packages,
        figure,
        **document_kwargs
    )

    return document

def _copy_if_needed(src, dst):
    try:
        shutil.copy(src, dst)
    except shutil.SameFileError:
        pass

def _move_if_needed(src, dst):
    try:
        shutil.move(src, dst)
    except shutil.SameFileError:
        pass

def _move_folder_or_move_files_if_destination_folder_exists(src, dst):
    try:
        _move_if_needed(src, dst)
    except shutil.Error:
        files = [file for file in next(os.walk(src))[2]]
        [_move_if_needed(file, dst) for file in files]

def _latex_valid_basename(filepath):
    basename = os.path.basename(filepath)
    return basename.replace(' ', '_').replace('/','_').replace('%','pct')


from statsmodels import api as sm

from .tools import _to_list_if_str, _to_list_if_tuple
from .fe import fixed_effects_reg_df_and_cols_dict, extract_all_dummy_cols_from_dummy_cols_dict
from .interact import create_interaction_variables, delete_interaction_variables, _collect_variables_from_interaction_tuples
from dero.reg.lag import create_lagged_variables, _convert_variable_names, _convert_interaction_tuples, \
    _set_lag_variables


def reg(df, yvar, xvars, robust=True, cluster=False, cons=True, fe=None, interaction_tuples=None,
        num_lags=0, lag_variables='xvars', lag_period_var='Date', lag_id_var='TICKER'):
    """
    Returns a fitted regression. Takes df, produces a regression df with no missing among needed
    variables, and fits a regression model. If robust is specified, uses heteroskedasticity-
    robust standard errors. If cluster is specified, calculated clustered standard errors
    by the given variable.

    Note: only specify at most one of robust and cluster.

    Required inputs:
    df: pandas dataframe containing regression data
    yvar: str, column name of outcome y variable
    xvars: list of strs, column names of x variables for regression

    Optional inputs:
    robust: bool, set to True to use heterskedasticity-robust standard errors
    cluster: False or str, set to a column name to calculate standard errors within clusters
             given by unique values of given column name
    cons: bool, set to False to not include a constant in the regression
    fe: None or str or list of strs. If a str or list of strs is passed, uses these categorical
    variables to construct dummies for fixed effects.
    interaction_tuples: tuple or list of tuples of column names to interact and include as xvars
    num_lags: int, Number of periods to lag variables. Setting to other than 0 will activate lags
    lag_variables: 'all', 'xvars', or list of strs of names of columns to lag for regressions.
    lag_period_var: str, only used if lag_variables is not None. name of column which
                    contains period variable for lagging
    lag_id_var: str, only used if lag_variables is not None. name of column which
                    contains identifier variable for lagging

    Returns:
    If fe=None, returns statsmodels regression result
    if fe is not None, returns a tuple of (statsmodels regression result, dummy_cols_dict)
    """
    regdf, y, X, dummy_cols_dict, lag_variables = _create_reg_df_y_x_and_dummies(df, yvar, xvars, cluster=cluster,
                                                                  cons=cons, fe=fe,
                                                                  interaction_tuples=interaction_tuples, num_lags=num_lags,
                                                                  lag_variables=lag_variables, lag_period_var=lag_period_var,
                                                                  lag_id_var=lag_id_var)

    mod = sm.OLS(y, X)

    result = _estimate_handling_robust_and_cluster(regdf, mod, robust, cluster)

    # Only return dummy_cols_dict when fe is active
    if fe is not None:
        return result, dummy_cols_dict
    else:
        return result

def _create_reg_df_y_x_and_dummies(df, yvar, xvars, cluster=False, cons=True, fe=None, interaction_tuples=None,
                                   num_lags=0, lag_variables='xvars', lag_period_var='Date', lag_id_var='TICKER'):
    # Handle lags
    if num_lags != 0:
        lag_variables = _set_lag_variables(lag_variables, yvar, xvars)
        df = create_lagged_variables(df, lag_variables, id_col=lag_id_var, date_col=lag_period_var, num_lags=num_lags)
        reg_yvar, reg_xvars = _convert_variable_names(yvar, xvars, lag_variables, num_lags=num_lags)
        if interaction_tuples is not None:
            interaction_tuples = _convert_interaction_tuples(interaction_tuples, lag_variables, num_lags=num_lags)
    else:
        reg_yvar, reg_xvars = yvar, xvars

    fe = _set_fe(fe)
    interaction_tuples = _set_interaction_tuples(interaction_tuples)
    regdf, y, X, dummy_cols_dict = _get_reg_df_y_x(df, reg_yvar, reg_xvars, cluster, cons, fe, interaction_tuples)
    return regdf, y, X, dummy_cols_dict, lag_variables

def _post_reg_cleanup(df, num_lags, lag_variables):
    # Cleanup lags
    if num_lags != 0:
        df.drop(lag_variables, axis=1, inplace=True)

def _estimate_handling_robust_and_cluster(regdf, model, robust, cluster, **fit_kwargs):
    assert not (robust and cluster)  # need to pick one of robust or cluster

    if robust:
        return model.fit(cov_type='HC1', **fit_kwargs)

    if cluster:
        groups = regdf[cluster].unique().tolist()
        group_ints = regdf[cluster].apply(lambda x: groups.index(x))
        return model.fit(cov_type='cluster', cov_kwds={'groups': group_ints}, **fit_kwargs)

        ##### TODO: implement this code for multiple clustering - need to ensure cluster is list
        # cluster_cols = ['State', 'Date']
        # reg_df = df.dropna()
        # groups = list(df[cluster_cols].drop_duplicates().itertuples(index=False))
        # group_ints = df[cluster_cols].apply(lambda x: groups.index(tuple(x)), axis=1)
        # result = sm.OLS(df['Return'], df['random']).fit(cov_type='cluster', cov_kwds={'groups': group_ints})
        #####

    return model.fit()

def _get_reg_df_y_x(df, yvar, xvars, cluster, cons, fe, interaction_tuples):
    all_xvars = _collect_all_variables_from_xvars_and_interaction_tuples(xvars, interaction_tuples)
    regdf = _drop_missings_df(df, yvar, all_xvars, cluster, fe)
    y, X, dummy_cols_dict = _y_X_from_df(regdf, yvar, xvars, cons, fe, interaction_tuples)

    return regdf, y, X, dummy_cols_dict

def _drop_missings_df(df, yvar, xvars, cluster, fe):
    drop_set = [yvar] + xvars
    if cluster:
        drop_set += [cluster]
    if fe is not None:
        drop_set += fe

    return df.dropna(subset=drop_set)

def _y_X_from_df(regdf, yvar, xvars, cons, fe, interaction_tuples):

    if fe is not None:
        regdf, dummy_cols_dict = fixed_effects_reg_df_and_cols_dict(regdf, fe)
        model_xvars = xvars + extract_all_dummy_cols_from_dummy_cols_dict(dummy_cols_dict)
    else:
        dummy_cols_dict = None
        model_xvars = xvars.copy()

    if interaction_tuples:
        interaction_vars = create_interaction_variables(regdf, interaction_tuples)
        model_xvars += interaction_vars

    y = regdf[yvar]
    X = regdf.loc[:, model_xvars]

    if cons:
        X = sm.add_constant(X)

    return y, X, dummy_cols_dict

def _set_fe(fe):
    if fe is None:
        return None
    else:
        return _to_list_if_str(fe)

def _set_interaction_tuples(interaction_tuples):
    if interaction_tuples is None:
        return []
    else:
        return _to_list_if_tuple(interaction_tuples)

def _collect_all_variables_from_xvars_and_interaction_tuples(xvars, interaction_tuples):
    interaction_vars = _collect_variables_from_interaction_tuples(interaction_tuples)
    return list(set(xvars + interaction_vars))


def _is_normal_reg_str(reg_str):
    return reg_str in ('reg', 'normal', 'ols') or reg_str == None


__author__ = 'Nick'

import os
import json

def _get_dropbox_location(account_type='personal'):
    """
    Returns a string of the filepath of the Dropbox for this user

    :param account_type: str, 'business' or 'personal'
    """
    info_path = _get_dropbox_info_path()
    info_dict = _get_dictionary_from_path_to_json(info_path)
    return _get_dropbox_path_from_dictionary(info_dict, account_type)

def _get_dropbox_info_path():
    """
    Returns filepath of Dropbox file info.json
    """
    path = _create_dropbox_info_path('APPDATA')
    if path:
        return path
    return _create_dropbox_info_path('LOCALAPPDATA')

def _create_dropbox_info_path(appdata_str):
    r"""
    Looks up the environment variable given by appdata_str and combines with \Dropbox\info.json

    Then checks if the info.json exists at that path, and if so returns the filepath, otherwise
    returns False
    """
    path = os.path.join(os.environ[appdata_str], r'Dropbox\info.json')
    if os.path.exists(path):
        return path
    return False

def _get_dictionary_from_path_to_json(info_path):
    """
    Loads a json file and returns as a dictionary
    """
    with open(info_path, 'r') as f:
        text = f.read()

    return json.loads(text)

def _get_dropbox_path_from_dictionary(info_dict, account_type):
    """
    Returns the 'path' value under the account_type dictionary within the main dictionary
    """
    return info_dict[account_type]['path']


filepath = _get_dropbox_location()

if __name__ == '__main__':
    print(filepath)




def select_models(reg_list, keepnum, xvars):
    """
    Takes a list of fitted regression models and selects among them based on adjusted R-Squared. For each
    number of variables involved in the regressions, keepnum with the highest R-squareds will be kept.

    For example, if reg_list contains 3 regressions with two variables and 6 regressions with three variables,
    and keepnum is 2, will return a list of four regressions, 2 with two variables and 2 with three variables.
    """
    outlist = []
    for i in range(1, len(xvars) + 1):
        reg_list_match = [reg for reg in reg_list if reg.df_model == i] #select models with this many variables
        try:
            r2_min = sorted([reg.rsquared_adj for reg in reg_list_match])[-keepnum] #gets keepnumth highest r2
        except IndexError: #should happen once there are less models run than keepnum (i.e. with all xvars)
            r2_min = sorted([reg.rsquared_adj for reg in reg_list_match])[0] #gets lowest r2 (keep all)
        outlist += [reg for reg in reg_list_match if reg.rsquared_adj >= r2_min]
    return outlist



from dero.latex.figure.models.subfigure import Subfigure
from dero.latex.models import Item
from dero.latex.models.caption import Caption
from dero.latex.models.label import Label
from dero.latex.logic.builder import build_content

class Figure(Item):
    name = 'figure'

    def __init__(self, subfigures: [Subfigure], caption=None, label=None, centering=True, position_str=None):
        self.subfigures = subfigures
        self.caption = Caption(caption) if caption else None
        self.label = Label(label) if label else None

        content = build_content(
            self.subfigures,
            caption=self.caption,
            label=self.label,
            centering=centering,
            position_str=position_str
        )

        super().__init__(self.name, content)

    def __repr__(self):
        return f'<Figure(subfigures={self.subfigures}, caption={self.caption})>'


from dero.latex.texgen import _begin_str, _end_str

class Environment:

    def __init__(self, name, modifiers=None):
        self.name = name

        self._begin = _begin_str(name)
        if modifiers:
            self._begin += modifiers

        self._end = _end_str(name)

    def __repr__(self):
        return f'<Environment(name={self.name})>'

    def wrap(self, other):
        return '\n'.join([self._begin, other, self._end])



from .iter import reg_for_each_combo, reg_for_each_xvar_set, reg_for_each_combo_select_and_produce_summary, \
    reg_for_each_xvar_set_and_produce_summary, reg_for_each_yvar_and_produce_summary
from .reg import reg
from .summarize import produce_summary
from .select import select_models
from .args import RegressionSetArgs





# from dero.latex.tools import date_time_move_latex
#
#
# def latex_equations_to_pdf(latex_list, directory, name='Equations', below_text=None,
#                            math_size=18, text_size=14, title=None, para_space='1em',
#                           inline=False, as_document=True):
#
#     # Header and footer are needed to create a standalone document using the equations.
#     # If as_document=False, header and footer will not be used.
#     headers = [r'\documentclass[{} pt]{{extarticle}}'.format(text_size),
#                #First size is text size, second is math size, third is script size,
#                #fourth is scriptscript size
#                r'\DeclareMathSizes{{{0}}}{{{1}}}{{{2}}}{{{3}}}'.format(
#                     text_size, math_size, script_size, scriptscript_size),
#                r'\usepackage{amsmath}',
#                r'\usepackage{breqn}',
#                r'\usepackage[margin=0.3in]{geometry}',
#               r'\author{Nick DeRobertis}' ,r'\begin{document}', r'\setlength{{\parskip}}{{{}}}'.format(para_space)]
#     footers = [r'\end{document}']
#
#     name_tex = name + '.tex'
#     file_path = os.path.join(directory, name_tex)
#
#     # Actually write to file
#     with open(file_path, 'w') as f:
#         if as_document:
#             f.write('\n'.join(headers) + '\n')
#         [f.write(surround_char_beg + '{}'.format(line) + surround_char_end + '\n\n') for line in latex_list]
#         if below_text:
#             f.write('\n' + below_text + '\n')
#         if as_document:
#             f.write('\n'.join(footers))
#
#     os.chdir(directory)
#     # Only create pdf if we are creating a standalone document
#     if as_document:
#         os.system('pdflatex ' + '"' + name_tex + '"') #create pdf
#     date_time_move_latex(name, directory, 'Equations')
#
# def _usepackage


from .ext_statsmodels import summary_col
import pandas as pd

from .fe import add_fixed_effects_rows
from .fe.tools import extract_all_dummy_cols_from_dummy_cols_dict_list, \
    extract_all_fe_names_from_dummy_cols_dict_list
from .controls import suppress_controls_in_summary_df



def produce_summary(reg_sets, stderr=False, float_format='%0.1f', regressor_order=[],
                    suppress_other_regressors=False, model_names=None):
    """

    :param reg_sets:
    :type reg_sets:
    :param stderr:
    :type stderr:
    :param float_format:
    :type float_format:
    :param regressor_order:
    :type regressor_order:
    :param suppress_other_regressors: True for when using regressor_order to suppress coefficients
        that are not in regressor_order into "Controls: Yes". False to keep coefficients
    :type suppress_other_regressors: bool
    :param model_names: If a collection is passed, will be used as column names in summary table.
    :type model_names: None, list, tuple
    :return:
    :rtype:
    """

    _check_produce_summary_inputs(regressor_order, suppress_other_regressors, model_names, len(reg_sets))

    # If not fixed effects, dummy_col_dicts will be None
    reg_list, dummy_col_dicts = _extract_result_list_and_dummy_dicts(reg_sets)

    info_dict = {'N': lambda x: "{0:d}".format(int(x.nobs)),
                                  # 'R2': lambda x: "{:.2f}".format(x.rsquared),
                                  'Adj-R2': lambda x: "{:.2f}".format(x.rsquared_adj)}

    summ = summary_col(reg_list, stars=True, float_format=float_format,
                       regressor_order=regressor_order,
                       info_dict=info_dict)

    # Handle fe - remove individual fe cols and replace with e.g. Industry Fixed Effects No, Yes, Yes
    if dummy_col_dicts: #if fixed effects
        split_rows = [var for var in info_dict]
        _remove_fe_cols_replace_with_fixed_effect_yes_no_lines(summ, dummy_col_dicts, split_rows)

    # Handle dropping of unimportant coefficients and replacing with Controls: Yes or No
    if suppress_other_regressors:
        summ.tables[0] = suppress_controls_in_summary_df(summ.tables[0], regressor_order, dummy_col_dicts,
                                                         info_dict)

    if not stderr:
        summ.tables[0].drop('', axis=0, inplace=True)  # drops the rows containing standard errors

    # Change const to Intercept in output
    summ.tables[0].index = [col if col != 'const' else 'Intercept' for col in summ.tables[0].index]

    if model_names:
        summ.tables[0].columns = model_names

    return summ

def _remove_fe_cols_replace_with_fixed_effect_yes_no_lines(summ, dummy_col_dicts, split_rows):
    """
    Note: inplace
    """
    # split into dataframe of variables and dataframe of N, R^2, etc.
    var_df, split_df = _get_var_df_and_non_var_df(summ.tables[0], split_rows=split_rows)
    # get name of all fixed effect variables
    all_cols_to_remove = extract_all_dummy_cols_from_dummy_cols_dict_list(dummy_col_dicts)
    # remove fixed effect coefs and stderrs
    var_df = _drop_variables_from_reg_summary_df(var_df, all_cols_to_remove)

    # construct a single dict where the keys are names of fixed effects and the values are lists of booleans for
    # whether the fixed effect was used
    fe_dict = _multiple_model_fe_dict_from_dummy_col_dict_list(dummy_col_dicts)

    # Add yes no row
    var_df = add_fixed_effects_rows(var_df, fe_dict)

    # Recombine with n, R^2, etc. and
    summ.tables[0] = pd.concat([var_df, split_df], axis=0)

def _extract_result_list_and_dummy_dicts(result_sets):
    plain_results = []
    dummy_dicts = []
    for ambiguous_result in result_sets:
        # This is the case where fe has been passed, and we have a dummy_col_dict
        if isinstance(ambiguous_result, tuple):
            plain_results.append(ambiguous_result[0])
            dummy_dicts.append(ambiguous_result[1])
        # No fe passed, just plain result
        else:
            plain_results.append(ambiguous_result)
            dummy_dicts.append(None)  # keep order by appending None

    return plain_results, dummy_dicts


def _multiple_model_fe_dict_from_dummy_col_dict_list(dummy_col_dict_list):
    fixed_effect_rows = extract_all_fe_names_from_dummy_cols_dict_list(dummy_col_dict_list)
    out_dict = {fe_name: [] for fe_name in fixed_effect_rows}
    for dummy_col_dict in dummy_col_dict_list:
        for fe_name in fixed_effect_rows:
            if (not dummy_col_dict) or (fe_name not in dummy_col_dict):
                out_dict[fe_name].append(False)
            else:
                out_dict[fe_name].append(True)

    return out_dict


def _drop_variables_from_reg_summary_df(df, dropvars):
    # Find variables to be kept
    keepvars = [var for var in df.index if var not in dropvars and var != '']

    # Create column identifying row as an estimate or standard error
    df['type'] = ['estimate', 'stderr'] * int(len(df.index) / 2)

    # Create column identifying variable name of row (no spaces)
    df['regressor'] = [i for sublist in [[j] * 2 for j in df.index[0::2]] for i in sublist]

    # Create a column of the original location of the row (will be sorting index, need to get original
    # order back later)
    df['idx'] = [i for i in range(len(df))]

    # Create the multi-index for slicing
    df = df.set_index(['regressor', 'type'])
    df = df.sort_index()

    # Slice on the chosen regressors, reset the index to delete a column later
    df = df.loc[keepvars].reset_index()
    df = df.sort_values('idx')

    # Set value of index back to original - which had blanks for stderrs
    df.loc[df['type'] == 'stderr', 'regressor'] = ''

    # Delete the type column
    df.drop(['type', 'idx'], axis=1, inplace=True)

    # Reindex the dataframe on the regressor
    df = df.set_index(['regressor'])

    # Get rid of name on index
    df.index.name = None

    return df

def _get_var_df_and_non_var_df(df, split_rows=['N', 'R2', 'Adj-R2']):
    """
    Splits rows containing N, R2, Adj-R2 into separate dataframe
    """
    other_data_mask = df.index.isin(split_rows)
    return df.loc[~other_data_mask], df.loc[other_data_mask]

def _check_produce_summary_inputs(regressor_order, supress_other_regressors, model_names, num_models):
    if (regressor_order == []) & (supress_other_regressors):
        raise ValueError('must pass regressors to regressor_order to suppress other regressors')

    if model_names and (len(model_names) != num_models):
        raise ValueError(f'must pass model_names of equal length to num models. Have {len(model_names)} names and {num_models} models.')


import numpy as np
import pandas as pd

from statsmodels.iolib.summary2 import lrange, reduce, zip, range, _col_params, \
    _make_unique, _col_info, Summary


def summary_col(results, float_format='%.4f', model_names=[], stars=False,
                info_dict=None, regressor_order=[]):
    """

    NOTE: the purpose of recreating the summary_col function is becuase of issue
    #3767 in statsmodels:
    https://github.com/statsmodels/statsmodels/issues/3767
    Which can cause coefficients to become mismatched when using regressor_order.
    I have patched the issue by replacing np.unique with pd.unique.

    TODO: Once the issue is patched, replace this function with simply a reference
    to summary_col from statsmodels


    Summarize multiple results instances side-by-side (coefs and SEs)

    Parameters
    ----------
    results : statsmodels results instance or list of result instances
    float_format : string
        float format for coefficients and standard errors
        Default : '%.4f'
    model_names : list of strings of length len(results) if the names are not
        unique, a roman number will be appended to all model names
    stars : bool
        print significance stars
    info_dict : dict
        dict of lambda functions to be applied to results instances to retrieve
        model info. To use specific information for different models, add a
        (nested) info_dict with model name as the key.
        Example: `info_dict = {"N":..., "R2": ..., "OLS":{"R2":...}}` would
        only show `R2` for OLS regression models, but additionally `N` for
        all other results.
        Default : None (use the info_dict specified in
        result.default_model_infos, if this property exists)
    regressor_order : list of strings
        list of names of the regressors in the desired order. All regressors
        not specified will be appended to the end of the list.
    """

    if not isinstance(results, list):
        results = [results]

    cols = [_col_params(x, stars=stars, float_format=float_format) for x in
            results]

    # Unique column names (pandas has problems merging otherwise)
    if model_names:
        colnames = _make_unique(model_names)
    else:
        colnames = _make_unique([x.columns[0] for x in cols])
    for i in range(len(cols)):
        cols[i].columns = [colnames[i]]

    merg = lambda x, y: x.merge(y, how='outer', right_index=True,
                                left_index=True)
    summ = reduce(merg, cols)

    if regressor_order:
        varnames = summ.index.get_level_values(0).tolist()
        ordered = [x for x in regressor_order if x in varnames]
        unordered = [x for x in varnames if x not in regressor_order + ['']]
        order = ordered + list(np.unique(unordered))

        f = lambda idx: sum([[x + 'coef', x + 'stde'] for x in idx], [])
        summ.index = f(pd.unique(varnames))
        summ = summ.reindex(f(order))
        summ.index = [x[:-4] for x in summ.index]

    idx = pd.Series(lrange(summ.shape[0])) % 2 == 1
    summ.index = np.where(idx, '', summ.index.get_level_values(0))

    # add infos about the models.
    if info_dict:
        cols = [_col_info(x, info_dict.get(x.model.__class__.__name__,
                                           info_dict)) for x in results]
    else:
        cols = [_col_info(x, getattr(x, "default_model_infos", None)) for x in
                results]
    # use unique column names, otherwise the merge will not succeed
    for df , name in zip(cols, _make_unique([df.columns[0] for df in cols])):
        df.columns = [name]
    merg = lambda x, y: x.merge(y, how='outer', right_index=True,
                                left_index=True)
    info = reduce(merg, cols)
    dat = pd.DataFrame(np.vstack([summ, info]))  # pd.concat better, but error
    dat.columns = summ.columns
    dat.index = pd.Index(summ.index.tolist() + info.index.tolist())
    summ = dat

    summ = summ.fillna('')

    smry = Summary()
    smry.add_df(summ, header=True, align='l')
    smry.add_text('Standard errors in parentheses.')
    if stars:
        smry.add_text('* p<.1, ** p<.05, ***p<.01')

    return smry


from dero.latex.texgen import _centering_str

def build_content(items, caption=None, label=None, centering=True, position_str=None):
    elements = [
        position_str,
        _centering_str() if centering else None,
        *[str(item) for item in items],
        str(caption) if caption else None,
        str(label) if label else None
    ]

    valid_elements = [elem for elem in elements if elem]

    return _build(valid_elements)

def _build(elements):
    return '\n'.join(elements)


def _begin_str(str_):
    return rf'\begin{{{str_}}}'


def _end_str(str_):
    return fr'\end{{{str_}}}'


def _include_graphics_str(filepath, width=r'\linewidth'):
    return rf'\includegraphics[width={width}]{{{filepath}}}'

def _basic_item_str(item_name, contents):
    return rf'\{item_name}{{{contents}}}'

def _centering_str():
    return r'\centering'

def _document_class_str():
    return r'\documentclass{article}'


def _usepackage_str(str_, modifier_str=None):
    if modifier_str:
        full_modifier_str = f'[{modifier_str}]'
    else:
        full_modifier_str = ''
    return rf'\usepackage{full_modifier_str}{{{str_}}}'